/*
Copyright 2016 Karolos Antoniadis, Peva Blanchard, Rachid Guerraoui and Julien Stainer <first.last@epfl.ch>

This file is part of CoRNG.

CoRNG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CoRNG is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CoRNG.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <linux/module.h>
#include <linux/syscalls.h>
#include <asm/syscalls.h>
#include <asm/irq.h>
#include <linux/sched.h>
#include <linux/unistd.h>
#include <linux/cpuset.h>
#include <linux/cpumask.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/kthread.h>
#include <linux/types.h>
#include <linux/semaphore.h>
#include <linux/completion.h>
#include <linux/hw_random.h>
#include <linux/random.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/uaccess.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/vmalloc.h>

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Concurrency Observation (CoOBS)");

#define KERNEL_SYSCALLs

#define CO0 0
#define CO1 2

//#define NUMA 
#define NUMA_NODE 0

// memory barrier
#define USEBAR

//#define TWOCL

/*
 * Data structures
 */

typedef struct tdata_str {
  uint64_t *m;
  uint64_t *y;
  uint64_t *o;
  struct completion comp;
  uint64_t nbrounds;
} tdata;






/* cache line flush, as found in:
 * http://stackoverflow.com/questions/20600798/clflush-not-flushing-the-instruction-cache
 *
 * TODO: Inline clflush directly in thread job.
 */
void __clflush(volatile void *p) { asm volatile("clflush (%0)" ::"r"(p)); }


/*
 * CoOBS thread job
 */

static int tjob(void *td) {
	int cpu = get_cpu(); // reserves cpu, prevents context switches
	//local_irq_disable(); // interrupts
	// printk(KERN_INFO "CoRNG - cpu taken: %d\n", cpu);

	tdata *d = (tdata *)td;
	uint64_t n;

	if (!d->o) {
		printk(KERN_ALERT "CoRNG - %p, Didn't allocate any memory", td);
		return -1;
	}

	for (n = 0; n < d->nbrounds; n++) {
		*(d->m) = n;
		asm volatile("" ::: "memory");
#ifdef USEBAR
		__sync_synchronize();
#endif
		__clflush(d->m);
		d->o[n] = *(d->y);
		asm volatile("" ::: "memory");
		__clflush(d->y);
	}

	complete(&(d->comp));
	//local_irq_enable();
	put_cpu();

	return 0;
}







/*
 * CoOBS open method
 * TODO: Put a lock to prevent two programs opening /dev/coobs at the same time.
 */

static int coobs_open(struct inode *inode, struct file *file) {
	printk(KERN_INFO "CoOBS - open.\n");
	return 0; 
}







/*
 * CoOBS release method
 */

static int coobs_release(struct inode *inode, struct file *file) { 
	printk(KERN_INFO "CoOBS - release.\n");
	return 0; 
}








/*
 * CoOBS read method
 */

static ssize_t coobs_read(struct file *file, char __user *buff, size_t size, loff_t *off) { 
	uint64_t i, j, k;

	/*
	 * CoOBS
	 *
	 * TODO: need 64B alignment
	 */

	tdata td[2];
	uint64_t nbrounds = size; /* WARNING: number of rounds should equal the number of bytes of the (unclean) compact trace. */ 
	uint64_t *m;

//	printk(KERN_ALERT "nbrounds : %d\n",nbrounds);

#ifndef NUMA
	m = (uint64_t*) kmalloc(16*64, GFP_KERNEL);
#else
	m = (uint64_t*) kmalloc_node(16*64, GFP_KERNEL, NUMA_NODE);
#endif
	if (m == NULL) {
		printk(KERN_ALERT "CoOBS - coobs_read(): couldn't allocate memory: m = kmalloc...(16*64...).\n");
		return -1;
	} else 	printk(KERN_INFO "CoOBS - coobs_read(): alignment &m[0] mod 64 == %lu.\n", ((unsigned long) &m[0])%64);  

	struct task_struct *th_st0, *th_st1;

	m[0] = ULONG_MAX;
	m[1] = ULONG_MAX;
	m[8] = ULONG_MAX;

	for (i=0; i<2; i++) {
#ifndef TWOCL
		td[i].m = &m[i];
		td[i].y = &m[1-i];
#else
		td[i].m = &m[8*i];
		td[i].y = &m[8*(1-i)];
#endif
		td[i].o = (uint64_t *)vmalloc(nbrounds*sizeof(uint64_t));
		if (td[i].o == NULL) {
			for (j=0; j<i; j++)
				if (td[j].o) vfree(td[j].o);
			kfree(m);
			printk(KERN_ALERT "CoOBS - coobs_read(): couldn't allocate memory: td[%d].o\n",i);
			return -1;
		}
		init_completion(&(td[i].comp));
		td[i].nbrounds = nbrounds;
	}

	th_st0 = kthread_create(tjob, (void*)(&td[0]), "thread 0");
	kthread_bind(th_st0, CO0);
	th_st1 = kthread_create(tjob, (void*)(&td[1]), "thread 0");
	kthread_bind(th_st1, CO1);

	wake_up_process(th_st0);
	wake_up_process(th_st1);

	if (th_st0)
		printk(KERN_INFO "CoOBS - thread 0 running on core %d.\n", task_cpu(th_st0));
	else {
		for (j=0; j<2; j++)
			if (td[j].o) vfree(td[j].o);
		kfree(m);
		printk(KERN_ALERT "CoOBS - thread 0 creation failed.\n", task_cpu(th_st0));
		return -1; 
	}
	if (th_st1)
		printk(KERN_INFO "CoOBS - thread 1 running on core %d.\n", task_cpu(th_st1));
	else {
		for (j=0; j<2; j++)
			if (td[j].o) vfree(td[j].o);
		kfree(m);
		printk(KERN_ALERT "CoOBS - thread 1 creation failed.\n", task_cpu(th_st1));
		return -1; 
	}

	wait_for_completion(&(td[0].comp));
	printk(KERN_INFO "CoOBS - thread 0 completed.\n");
	wait_for_completion(&(td[1].comp));
	printk(KERN_INFO "CoOBS - thread 1 completed.\n");


	/*
	 * CoREC
	 */

	char *mc = (char*) vmalloc(2*nbrounds*sizeof(char));
	if (mc == NULL) {
		for (j=0; j<2; j++)
			if (td[j].o) vfree(td[j].o);
		kfree(m);
		printk(KERN_ALERT "CoOBS - coobs_read(): couldn't allocate memory.\n");
		return -1; 
	}

	for (k=0; k<2*nbrounds; k++)
		mc[k] = -1;


	/* offline detection, filling mc from the threads' arrays of read values */
	uint64_t n0 = 0;
	uint64_t n1 = 0;
	while (n0 < nbrounds && n1 < nbrounds) {
		if (td[0].o[n0] == ULONG_MAX) {
			mc[n0 + n1] = 0; /* w0 r0 */
			n0++;
		} else {
			if (td[1].o[n1] == ULONG_MAX) {
				mc[n0 + n1] = 1; /* w1 r1 */
				n1++;
			} else {
				if (td[0].o[n0] < n1) {
					mc[n0 + n1] = 0; /* w0 r0 */
					n0++;
				} else {
					if (td[1].o[n1] < n0) {
						mc[n0 + n1] = 1; /* w1 r1 */
						n1++;
					} else {
						if (td[0].o[n0] == n1 || td[1].o[n1] == n0) {
							mc[n0 + n1] = 4; /* w0 || w1 */
							while (td[1].o[n1] > n0 || td[0].o[n0] > n1) {
								if (td[1].o[n1] > n0) {
									mc[n0 + n1 + 1] = 2; /* r0 w0 */
									n0++;
								} else {
									mc[n0 + n1 + 1] = 3; /* r1 w1 */
									n1++;
								}
							}
							mc[n0 + n1 + 1] = 5; /* r0 || r1 */
							n0++;
							n1++;
						} else {
							vfree(mc);
							for (j=0; j<2; j++)
								if (td[j].o) vfree(td[j].o);
							kfree(m);
							printk(KERN_ALERT "CoRNG - coobs_read(): reconstruction error.\n");
							return -1;
						}
					}
				}
			}
		}
	}
	while (n0 < nbrounds) {
		mc[n0+n1] = 0; /* w0 r0 */
		n0++;
	}
	while (n1 < nbrounds) {
		mc[n0+n1] = 1; /* w1 r1 */
		n1++;
	}

	uint64_t tracelen = 2*nbrounds; /* should equal n0+n1 */
	if (tracelen % 2 != 0) tracelen--; /* should never happen */
	if (tracelen == 0) {
		vfree(mc);
		for (j=0; j<2; j++)
			if (td[j].o) vfree(td[j].o);
		kfree(m);
		printk(KERN_ALERT "CoOBS - coobs_read(): reconstruction error: zero length.\n");
		return -1;
	}

	/*
	 * Copy to user
	 */
	uint64_t compact_size = tracelen >> 1; /* divide by 2 */
	compact_size = (compact_size > size) ? size : compact_size; /* min between requested size and current size */
	char *compact = (char*) vmalloc(compact_size*sizeof(char));
	if (compact == NULL) {
		vfree(mc);
		for (j=0; j<2; j++)
			if (td[j].o) vfree(td[j].o);
		kfree(m);
		printk(KERN_ALERT "CoOBS - coobs_read(): couldn't allocate memory: compact = vmalloc(...).\n");
		return -1;
	}

	for (i=0; i<compact_size; i++) {
		char a = (char) mc[2*i];
		char b = (char) mc[2*i+1];
		char mask = 15;
		compact[i] = (a & mask) + (b << 4); 
	}
	
	ssize_t result_size;
	int not_copied_bytes = copy_to_user(buff, compact, compact_size);
	result_size = compact_size - not_copied_bytes;

	/*
	 * Free memory
	 */

	vfree(compact);
	vfree(mc);
	for (j=0; j<2; j++)
		if (td[j].o) vfree(td[j].o);
	kfree(m);

	//clean_observation(td);

	return result_size;
}


static struct file_operations coobs_fops = {
	.owner = THIS_MODULE,
	.open = coobs_open,
	.read = coobs_read,
	.release = coobs_release
};

static struct miscdevice coobs_device = {
    .minor = MISC_DYNAMIC_MINOR, 
    .name = "coobs", 
    .fops = &coobs_fops
};

static int __init coobs_init(void) {
	printk(KERN_INFO "CoOBS - module init.\n");
	misc_register(&coobs_device);
	printk(KERN_INFO "CoOBS - coobs device has minor number: %d\n", coobs_device.minor);
	return 0;
}

static void __exit coobs_cleanup(void) {
	printk(KERN_INFO "CoOBS - module exit.\n");
	misc_deregister(&coobs_device);
}


module_init(coobs_init);
module_exit(coobs_cleanup);
